#!/bin/bash
#
# all main functions to generate a PKI certificate chain
# 

#
# Set the CA variables
#
pki_func_init() {
  if [[ -n $1 ]] || [[ -n $2 ]] || [[ -n $3 ]]; then
    FQ_CA_CERT=$1
    FQ_CA_KEYS=$2
    CNF_PATH=$3
    APP_INIT=1
  else
    APP_INIT=0
  fi
}

#
# print text wrapped in a block
#
echo_block() {
  echo
  echo "***** ***** ***** *****"
  echo $1
  echo "***** ***** ***** *****"
}

# 
# Grab the latest serial # from the file, auto-increment
# 
get_serial() {
  SERIAL=`head "cfg/SERIAL"`
  if [[ -z $SERIAL ]]; then
    SERIAL=11111
    echo_block "WARN: no file 'SERIAL' found, using default 11111 as the serial # for CA"
  fi
}

# ***** ***** ***** ***** *****
# 
# CERTIFICATE AUTHORITY (CA)
# 
# ***** ***** ***** ***** *****
# This function will generate a CA Intermediate
# IN: UNIQ_ID_CA, SERIAL
#
gen_ca() {
  UNIQ_ID_CA=$1
  SERIAL=$2

  echo_block "Create CA (${UNIQ_ID_CA})"

  # encrypt the key
  #openssl genrsa -aes256 -out ca.keys.pem 4096
  #openssl genrsa -aes256 -password "pass:password" -out ca.keys.pem 4096 

  # key un-protected
  openssl genrsa -out "ca_${UNIQ_ID_CA}.keys.pem" 4096
  # 
  # Create Certificate   (valid for 10 years, after the entire chain of trust expires)
  openssl req -config $CNF_PATH/ca.cnf -new -x509 -sha256 -days 3650 -extensions v3_ca \
              -subj "/C=OO/O=ACME/CN=root.${UNIQ_ID_CA}" -set_serial ${SERIAL} \
              -key ca_${UNIQ_ID_CA}.keys.pem -out ca_${UNIQ_ID_CA}.crt.pem

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in ca_${UNIQ_ID_CA}.crt.pem > ca_${UNIQ_ID_CA}_cert.info.txt
}

#
# Create CA Intermediate PKI
# 
# 

#
# Generate a PKI chain
#   - the certificate chain is unique based on the serial #
#   - generate a new CA I
#   - generate server certificates
#   - generate client certificates
#
# INPUT:  BASE SERIAL #, LOOP NUM
# 
# Requires: FQ_CA_CERT, FQ_CA_KEYS
# 
ca-i_gen_pki() {
  CDD=`pwd`
  ORG_URL=$1
  SERIAL_O=$2
  NUM_CERTS=$(($3-1))

  # create unique directory
  UNIQ_ID_CAI="${SERIAL_O}.${ORG_URL}" 
  mkdir -p "distribution/ca_i_${UNIQ_ID_CAI}"
  cd "distribution/ca_i_${UNIQ_ID_CAI}"

  # Create CA Intermediate
  ca-i_gen_cert $ORG_URL $SERIAL_O

  # create directories, copy files, before generating client/server
  ca-i_create_shell

  __ca-i_gen_client

  __ca-i_gen_server

  # return to last path
  cd $CDD
}

# 
# Client Certificates
# 
__ca-i_gen_client() {
  # create directories
  mkdir -p clients/data
  mkdir -p clients/distro
  mkdir -p clients/docs
  cd clients
  for NUM in $(seq 0 $NUM_CERTS)
  do
    gen_client $ORG_URL $((SERIAL_O+NUM))
  done
  cd ..
}

# 
# Server Certificates
# 
__ca-i_gen_server() {
  # create directories
  mkdir -p servers/data
  mkdir -p servers/distro
  mkdir -p servers/docs
  cd servers
  for NUM in $(seq 0 $NUM_CERTS)
  do
    gen_server $ORG_URL $((SERIAL_O+NUM))
  done
  cd ..
}

# This function will generate a CA Intermediate
# 
# Requires:  CNF file, CA cert, CA key
# 
# IN: UNIQ_ID_CA, SERIAL
#
ca-i_gen_cert() {
  ORG_URL=$1 
  SERIAL=$2

  UNIQ_ID="${SERIAL}.${ORG_URL}"

  echo_block "Create CA Intermediate  (${UNIQ_ID})"

  openssl genrsa -out "ca_i_${UNIQ_ID}.keys.pem" 4096

  # Create Cert Signing Request (CSR)
  openssl req -config "${CNF_PATH}/ca.cnf" -new -sha256 \
              -subj "/C=OO/O=ACME/OU=ACME Intermediate/CN=${UNIQ_ID}" \
              -key "ca_i_${UNIQ_ID}.keys.pem" -out "ca_i_${UNIQ_ID}.csr.pem"

  # Create Certificate   (valid for ~2 years, after the entire chain of trust expires)
  # CA signs Intermediate
  openssl x509 -req -days 750 -extfile "${CNF_PATH}/ca.cnf" -extensions v3_ca_i \
               -CA $FQ_CA_CERT -CAkey $FQ_CA_KEYS -set_serial ${SERIAL} \
               -in "ca_i_${UNIQ_ID}.csr.pem" -out "ca_i_${UNIQ_ID}.crt.pem"

  # Package the Certificate Authority Certificates for distro  (windoze needs this)
  openssl pkcs12 -export -password "pass:password" -inkey "ca_i_${UNIQ_ID}.keys.pem" \
                 -name "CA Intermediate Mobile Provision" -certfile $FQ_CA_CERT \
                 -in "ca_i_${UNIQ_ID}.crt.pem" -out "ca_i_${UNIQ_ID}.p12"

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in "ca_i_${UNIQ_ID}.crt.pem" > "ca_i_${UNIQ_ID}.crt.info.txt"

  # create certifiate chain
  cat $FQ_CA_CERT "ca_i_${UNIQ_ID}.crt.pem" > "ca_cert-chain_${UNIQ_ID}.crts.pem"
}

#
# Copies all applcations to the Lifecycle package
#   organize the ca-i directory
#   order matters:  move these files last because they were copied above
#
ca-i_create_shell() {

  DEST_DIR="${CDD}/distribution/ca_i_${UNIQ_ID_CAI}"

  # client
  mkdir -p clients/cfg
  cp $CDD/res/libs/gen_client.sh  $DEST_DIR/clients/
  cp $CDD/res/libs/pki_funcs.sh   $DEST_DIR/clients/cfg
  cp $CDD/res/docs/README_C       $DEST_DIR/clients/README
  cp $CDD/res/docs/SERIAL         $DEST_DIR/clients/cfg/
  cp "${CDD}/cfg/${ORG_URL}.cnf"  $DEST_DIR/clients/cfg/
  # generated files
  cp $DEST_DIR/ca_i*.crt.pem      $DEST_DIR/clients/cfg/ca-i.crt.pem
  cp $DEST_DIR/ca_i*.keys.pem     $DEST_DIR/clients/cfg/ca-i.keys.pem
  cp $DEST_DIR/ca_cert-chain*.pem $DEST_DIR/clients/cfg/ca_cert-chain.crts.pem

  # server
  mkdir -p servers/cfg
  cp $CDD/res/libs/gen_server.sh  $DEST_DIR/servers/
  cp $CDD/res/libs/pki_funcs.sh   $DEST_DIR/servers/cfg/
  cp $CDD/res/docs/README_S       $DEST_DIR/servers/README
  cp $CDD/res/docs/SERIAL         $DEST_DIR/servers/cfg/
  cp "${CDD}/cfg/${ORG_URL}.cnf"  $DEST_DIR/servers/cfg/
  # generated files
  cp $DEST_DIR/ca_i*.crt.pem      $DEST_DIR/servers/cfg/ca-i.crt.pem
  cp $DEST_DIR/ca_i*.keys.pem     $DEST_DIR/servers/cfg/ca-i.keys.pem
  cp $DEST_DIR/ca_cert-chain*.pem $DEST_DIR/servers/cfg/ca_cert-chain.crts.pem

  # CA-I
  mkdir -p ca-i/data
  mkdir -p ca-i/docs
  mkdir -p ca-i/distro
  cp $CDD/res/docs/README_CAI       $DEST_DIR/README
  cp $CDD/ca_*/ca_*.crt.pem         $DEST_DIR/ca-i/data/
  cp $CDD/ca_*/ca_*.info.txt        $DEST_DIR/ca-i/docs/
  # generated files
  mv $DEST_DIR/ca_i*.pem            $DEST_DIR/ca-i/data/
  mv $DEST_DIR/ca_i*.info.txt       $DEST_DIR/ca-i/docs/
  mv $DEST_DIR/ca_i*.p12            $DEST_DIR/ca-i/distro
  mv $DEST_DIR/ca_cert-chain*.pem   $DEST_DIR/ca-i/distro
}

#
# Generate a Client Certificate
# IN: UNIQ_ID, UNIQ_ID_CAI, SERIAL
#
gen_client() {
  ORG_URL=$1 
  SERIAL=$2

  UNIQ_ID="${SERIAL}.${ORG_URL}"
  CERT_CHAIN="cfg/ca_cert-chain.crts.pem"

  echo_block "Generate Client Certificates  (${UNIQ_ID})"

  openssl genrsa -out "data/client_${UNIQ_ID}.keys.pem" 4096

  openssl req -new -key "data/client_${UNIQ_ID}.keys.pem" \
              -subj "/C=OO/O=ACME/OU=ACME Standard/CN=client_${UNIQ_ID}" \
              -out "data/client_${UNIQ_ID}.csr.pem"
  # CA Intermediate signs Client
  openssl x509 -req -days 365 \
               -CA "cfg/ca-i.crt.pem" -CAkey "cfg/ca-i.keys.pem" -set_serial ${SERIAL} \
               -in "data/client_${UNIQ_ID}.csr.pem" -out "data/client_${UNIQ_ID}.crt.pem"

  # Package the Certificates
  openssl pkcs12 -export -password "pass:password" -inkey "data/client_${UNIQ_ID}.keys.pem" \
                 -name "Client ${UNIQ_ID} VPN Certificate" -certfile $CERT_CHAIN -caname "client_${UNIQ_ID}@acme.xyz" \
                 -in "data/client_${UNIQ_ID}.crt.pem" -out "distro/client_${UNIQ_ID}.p12"

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in "data/client_${UNIQ_ID}.crt.pem" > "docs/client_${UNIQ_ID}.info.txt"
}

#
# Generate a Server Certificate
# IN: UNIQ_ID, UNIQ_ID_CA, SERIAL
#
gen_server() {
  ORG_URL=$1 
  SERIAL=$2

  UNIQ_ID="${SERIAL}.${ORG_URL}"
  CERT_CHAIN="cfg/ca_cert-chain.crts.pem"

  echo_block "Generate Server Certificates  (${UNIQ_ID})"

  openssl genrsa -out "data/server_${UNIQ_ID}.keys.pem" 4096

  openssl req -new -config "cfg/${ORG_URL}.cnf" -key "data/server_${UNIQ_ID}.keys.pem" \
              -subj "/C=OO/O=ACME/OU=ACME Standard/CN=${UNIQ_ID}" \
              -out "data/server_${UNIQ_ID}.csr.pem"

  # CA Intermediate signs Server
  openssl x509 -req -days 365 -extfile "cfg/${ORG_URL}.cnf" -extensions v3_server \
               -CA "cfg/ca-i.crt.pem" -CAkey "cfg/ca-i.keys.pem" -set_serial ${SERIAL} \
               -in "data/server_${UNIQ_ID}.csr.pem" -out "data/server_${UNIQ_ID}.crt.pem"

  # Package the Certificates
  openssl pkcs12 -export -password "pass:password" -inkey "data/server_${UNIQ_ID}.keys.pem" \
                 -name "Server ${UNIQ_ID} VPN Certificate" -certfile $CERT_CHAIN -caname "server_${UNIQ_ID}@acme.xyz" \
                 -in "data/server_${UNIQ_ID}.crt.pem" -out "distro/server_${UNIQ_ID}.p12"

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in "data/server_${UNIQ_ID}.crt.pem" > "docs/server_${UNIQ_ID}.crt.info.txt" 
}

